@@ -8,7 +8,6 @@ import struct
 import socket
 
 from . import dpkt
-from .decorators import deprecated
 from .compat import compat_ord
 
 
@@ -21,6 +20,8 @@ from .compat import compat_ord
 # Cease Subcodes - RFC 4486
 # NOPEER Community - RFC 3765
 # Multiprotocol Extensions - 2858
+# Advertisement of Multiple Paths in BGP - RFC 7911
+# BGP Support for Four-Octet Autonomous System (AS) Number Spac - RFC 6793
 
 # Message Types
 OPEN = 1
@@ -129,7 +130,7 @@ class BGP(dpkt.Packet):
     """Border Gateway Protocol.
 
     BGP is an inter-AS routing protocol.
-    See more about the BGP on \
+    See more about the BGP on
     https://en.wikipedia.org/wiki/Border_Gateway_Protocol
 
     Attributes:
@@ -151,7 +152,7 @@ class BGP(dpkt.Packet):
         elif self.type == UPDATE:
             self.data = self.update = self.Update(self.data)
         elif self.type == NOTIFICATION:
-            self.data = self.notifiation = self.Notification(self.data)
+            self.data = self.notification = self.Notification(self.data)
         elif self.type == KEEPALIVE:
             self.data = self.keepalive = self.Keepalive(self.data)
         elif self.type == ROUTE_REFRESH:
@@ -171,14 +172,14 @@ class BGP(dpkt.Packet):
 
         def unpack(self, buf):
             dpkt.Packet.unpack(self, buf)
-            l = []
+            l_ = []
             plen = self.param_len
             while plen > 0:
                 param = self.Parameter(self.data)
                 self.data = self.data[len(param):]
                 plen -= len(param)
-                l.append(param)
-            self.data = self.parameters = l
+                l_.append(param)
+            self.data = self.parameters = l_
 
         def __len__(self):
             return self.__hdr_len__ + sum(map(len, self.parameters))
@@ -231,44 +232,47 @@ class BGP(dpkt.Packet):
             # Withdrawn Routes
             wlen = struct.unpack('>H', self.data[:2])[0]
             self.data = self.data[2:]
-            l = []
+            l_ = []
             while wlen > 0:
                 route = RouteIPV4(self.data)
                 self.data = self.data[len(route):]
                 wlen -= len(route)
-                l.append(route)
-            self.withdrawn = l
+                l_.append(route)
+            self.withdrawn = l_
 
             # Path Attributes
             plen = struct.unpack('>H', self.data[:2])[0]
             self.data = self.data[2:]
-            l = []
+            l_ = []
             while plen > 0:
                 attr = self.Attribute(self.data)
                 self.data = self.data[len(attr):]
                 plen -= len(attr)
-                l.append(attr)
-            self.attributes = l
+                l_.append(attr)
+            self.attributes = l_
 
             # Announced Routes
-            l = []
+            l_ = []
             while self.data:
-                route = RouteIPV4(self.data)
+                if len(self.data) % 9 == 0:
+                    route = ExtendedRouteIPV4(self.data)
+                else:
+                    route = RouteIPV4(self.data)
                 self.data = self.data[len(route):]
-                l.append(route)
-            self.announced = l
+                l_.append(route)
+            self.announced = l_
 
         def __len__(self):
             return 2 + sum(map(len, self.withdrawn)) + \
-                   2 + sum(map(len, self.attributes)) + \
-                   sum(map(len, self.announced))
+                2 + sum(map(len, self.attributes)) + \
+                sum(map(len, self.announced))
 
         def __bytes__(self):
             return struct.pack('>H', sum(map(len, self.withdrawn))) + \
-                   b''.join(map(bytes, self.withdrawn)) + \
-                   struct.pack('>H', sum(map(len, self.attributes))) + \
-                   b''.join(map(bytes, self.attributes)) + \
-                   b''.join(map(bytes, self.announced))
+                b''.join(map(bytes, self.withdrawn)) + \
+                struct.pack('>H', sum(map(len, self.attributes))) + \
+                b''.join(map(bytes, self.attributes)) + \
+                b''.join(map(bytes, self.announced))
 
         class Attribute(dpkt.Packet):
             __hdr__ = (
@@ -371,12 +375,16 @@ class BGP(dpkt.Packet):
 
                 def unpack(self, buf):
                     self.data = buf
-                    l = []
+                    l_ = []
+                    as4 = len(self.data) == 6
                     while self.data:
-                        seg = self.ASPathSegment(self.data)
+                        if as4:
+                            seg = self.ASPathSegment4(self.data)
+                        else:
+                            seg = self.ASPathSegment(self.data)
                         self.data = self.data[len(seg):]
-                        l.append(seg)
-                    self.data = self.segments = l
+                        l_.append(seg)
+                    self.data = self.segments = l_
 
                 def __len__(self):
                     return sum(map(len, self.data))
@@ -392,12 +400,12 @@ class BGP(dpkt.Packet):
 
                     def unpack(self, buf):
                         dpkt.Packet.unpack(self, buf)
-                        l = []
+                        l_ = []
                         for i in range(self.len):
                             AS = struct.unpack('>H', self.data[:2])[0]
                             self.data = self.data[2:]
-                            l.append(AS)
-                        self.data = self.path = l
+                            l_.append(AS)
+                        self.data = self.path = l_
 
                     def __len__(self):
                         return self.__hdr_len__ + 2 * len(self.path)
@@ -408,6 +416,31 @@ class BGP(dpkt.Packet):
                             as_str += struct.pack('>H', AS)
                         return self.pack_hdr() + as_str
 
+                class ASPathSegment4(dpkt.Packet):
+                    __hdr__ = (
+                        ('type', 'B', 0),
+                        ('len', 'B', 0)
+                    )
+
+                    def unpack(self, buf):
+                        dpkt.Packet.unpack(self, buf)
+                        l_ = []
+                        for i in range(self.len):
+                            if len(self.data) >= 4:
+                                AS = struct.unpack('>I', self.data[:4])[0]
+                                self.data = self.data[4:]
+                                l_.append(AS)
+                        self.path = l_
+
+                    def __len__(self):
+                        return self.__hdr_len__ + 4 * len(self.path)
+
+                    def __bytes__(self):
+                        as_str = b''
+                        for AS in self.path:
+                            as_str += struct.pack('>I', AS)
+                        return self.pack_hdr() + as_str
+
             class NextHop(dpkt.Packet):
                 __hdr__ = (
                     ('ip', 'I', 0),
@@ -446,7 +479,7 @@ class BGP(dpkt.Packet):
 
                 def unpack(self, buf):
                     self.data = buf
-                    l = []
+                    l_ = []
                     while self.data:
                         val = struct.unpack('>I', self.data[:4])[0]
                         if (0x00000000 <= val <= 0x0000ffff) or (0xffff0000 <= val <= 0xffffffff):
@@ -454,8 +487,8 @@ class BGP(dpkt.Packet):
                         else:
                             comm = self.Community(self.data[:4])
                         self.data = self.data[len(comm):]
-                        l.append(comm)
-                    self.data = self.list = l
+                        l_.append(comm)
+                    self.data = self.list = l_
 
                 def __len__(self):
                     return sum(map(len, self.data))
@@ -486,12 +519,12 @@ class BGP(dpkt.Packet):
 
                 def unpack(self, buf):
                     self.data = buf
-                    l = []
+                    l_ = []
                     while self.data:
                         id = struct.unpack('>I', self.data[:4])[0]
                         self.data = self.data[4:]
-                        l.append(id)
-                    self.data = self.list = l
+                        l_.append(id)
+                    self.data = self.list = l_
 
                 def __len__(self):
                     return 4 * len(self.list)
@@ -512,20 +545,30 @@ class BGP(dpkt.Packet):
                     dpkt.Packet.unpack(self, buf)
 
                     # Next Hop
+                    hop_len = 4
+                    if self.afi == AFI_IPV6:
+                        hop_len = 16
+                    l_ = []
                     nlen = struct.unpack('B', self.data[:1])[0]
                     self.data = self.data[1:]
+                    # next_hop is kept for backward compatibility
                     self.next_hop = self.data[:nlen]
-                    self.data = self.data[nlen:]
+                    while nlen > 0:
+                        hop = self.data[:hop_len]
+                        l_.append(hop)
+                        self.data = self.data[hop_len:]
+                        nlen -= hop_len
+                    self.next_hops = l_
 
                     # SNPAs
-                    l = []
+                    l_ = []
                     num_snpas = struct.unpack('B', self.data[:1])[0]
                     self.data = self.data[1:]
                     for i in range(num_snpas):
                         snpa = self.SNPA(self.data)
                         self.data = self.data[len(snpa):]
-                        l.append(snpa)
-                    self.snpas = l
+                        l_.append(snpa)
+                    self.snpas = l_
 
                     if self.afi == AFI_IPV4:
                         Route = RouteIPV4
@@ -537,28 +580,28 @@ class BGP(dpkt.Packet):
                         Route = RouteGeneric
 
                     # Announced Routes
-                    l = []
+                    l_ = []
                     while self.data:
                         route = Route(self.data)
                         self.data = self.data[len(route):]
-                        l.append(route)
-                    self.data = self.announced = l
+                        l_.append(route)
+                    self.data = self.announced = l_
 
                 def __len__(self):
                     return self.__hdr_len__ + \
-                           1 + len(self.next_hop) + \
-                           1 + sum(map(len, self.snpas)) + \
-                           sum(map(len, self.announced))
+                        1 + sum(map(len, self.next_hops)) + \
+                        1 + sum(map(len, self.snpas)) + \
+                        sum(map(len, self.announced))
 
                 def __bytes__(self):
                     return self.pack_hdr() + \
-                           struct.pack('B', len(self.next_hop)) + \
-                           bytes(self.next_hop) + \
-                           struct.pack('B', len(self.snpas)) + \
-                           b''.join(map(bytes, self.snpas)) + \
-                           b''.join(map(bytes, self.announced))
+                        struct.pack('B', sum(map(len, self.next_hops))) + \
+                        b''.join(map(bytes, self.next_hops)) + \
+                        struct.pack('B', len(self.snpas)) + \
+                        b''.join(map(bytes, self.snpas)) + \
+                        b''.join(map(bytes, self.announced))
 
-                class SNPA(object):
+                class SNPA(dpkt.Packet):
                     __hdr__ = (
                         ('len', 'B', 0),
                     )
@@ -586,12 +629,12 @@ class BGP(dpkt.Packet):
                         Route = RouteGeneric
 
                     # Withdrawn Routes
-                    l = []
+                    l_ = []
                     while self.data:
                         route = Route(self.data)
                         self.data = self.data[len(route):]
-                        l.append(route)
-                    self.data = self.withdrawn = l
+                        l_.append(route)
+                    self.data = self.withdrawn = l_
 
                 def __len__(self):
                     return self.__hdr_len__ + sum(map(len, self.data))
@@ -659,6 +702,17 @@ class RouteIPV4(dpkt.Packet):
         return self.pack_hdr() + self.prefix[:(self.len + 7) // 8]
 
 
+class ExtendedRouteIPV4(RouteIPV4):
+    __hdr__ = (
+        ('path_id', 'I', 0),
+        ('len', 'B', 0),
+    )
+
+    def __repr__(self):
+        cidr = '%s/%d PathId %d' % (socket.inet_ntoa(self.prefix), self.len, self.path_id)
+        return '%s(%s)' % (self.__class__.__name__, cidr)
+
+
 class RouteIPV6(dpkt.Packet):
     __hdr__ = (
         ('len', 'B', 0),
@@ -736,16 +790,57 @@ class RouteEVPN(dpkt.Packet):
 
 
 __bgp1 = b'\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x00\x13\x04'
-__bgp2 = b'\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x00\x63\x02\x00\x00\x00\x48\x40\x01\x01\x00\x40\x02\x0a\x01\x02\x01\xf4\x01\xf4\x02\x01\xfe\xbb\x40\x03\x04\xc0\xa8\x00\x0f\x40\x05\x04\x00\x00\x00\x64\x40\x06\x00\xc0\x07\x06\xfe\xba\xc0\xa8\x00\x0a\xc0\x08\x0c\xfe\xbf\x00\x01\x03\x16\x00\x04\x01\x54\x00\xfa\x80\x09\x04\xc0\xa8\x00\x0f\x80\x0a\x04\xc0\xa8\x00\xfa\x16\xc0\xa8\x04'
-__bgp3 = b'\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x00\x79\x02\x00\x00\x00\x62\x40\x01\x01\x00\x40\x02\x00\x40\x05\x04\x00\x00\x00\x64\xc0\x10\x08\x00\x02\x01\x2c\x00\x00\x01\x2c\xc0\x80\x24\x00\x00\xfd\xe9\x40\x01\x01\x00\x40\x02\x04\x02\x01\x15\xb3\x40\x05\x04\x00\x00\x00\x2c\x80\x09\x04\x16\x05\x05\x05\x80\x0a\x04\x16\x05\x05\x05\x90\x0e\x00\x1e\x00\x01\x80\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x0c\x04\x04\x04\x00\x60\x18\x77\x01\x00\x00\x01\xf4\x00\x00\x01\xf4\x85'
-__bgp4 = b'\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x00\x2d\x01\x04\x00\xed\x00\x5a\xc6\x6e\x83\x7d\x10\x02\x06\x01\x04\x00\x01\x00\x01\x02\x02\x80\x00\x02\x02\x02\x00'
+__bgp2 = (
+    b'\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x00\x63\x02\x00\x00\x00\x48\x40\x01'
+    b'\x01\x00\x40\x02\x0a\x01\x02\x01\xf4\x01\xf4\x02\x01\xfe\xbb\x40\x03\x04\xc0\xa8\x00\x0f\x40\x05\x04'
+    b'\x00\x00\x00\x64\x40\x06\x00\xc0\x07\x06\xfe\xba\xc0\xa8\x00\x0a\xc0\x08\x0c\xfe\xbf\x00\x01\x03\x16'
+    b'\x00\x04\x01\x54\x00\xfa\x80\x09\x04\xc0\xa8\x00\x0f\x80\x0a\x04\xc0\xa8\x00\xfa\x16\xc0\xa8\x04'
+)
+__bgp3 = (
+    b'\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x00\x79\x02\x00\x00\x00\x62\x40\x01'
+    b'\x01\x00\x40\x02\x00\x40\x05\x04\x00\x00\x00\x64\xc0\x10\x08\x00\x02\x01\x2c\x00\x00\x01\x2c\xc0\x80'
+    b'\x24\x00\x00\xfd\xe9\x40\x01\x01\x00\x40\x02\x04\x02\x01\x15\xb3\x40\x05\x04\x00\x00\x00\x2c\x80\x09'
+    b'\x04\x16\x05\x05\x05\x80\x0a\x04\x16\x05\x05\x05\x90\x0e\x00\x1e\x00\x01\x80\x0c\x00\x00\x00\x00\x00'
+    b'\x00\x00\x00\x0c\x04\x04\x04\x00\x60\x18\x77\x01\x00\x00\x01\xf4\x00\x00\x01\xf4\x85'
+)
+__bgp4 = (
+    b'\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x00\x2d\x01\x04\x00\xed\x00\x5a\xc6'
+    b'\x6e\x83\x7d\x10\x02\x06\x01\x04\x00\x01\x00\x01\x02\x02\x80\x00\x02\x02\x02\x00'
+)
 
 # BGP-EVPN type 1-4 packets for testing.
-__bgp5 = b'\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x00\x60\x02\x00\x00\x00\x49\x40\x01\x01\x00\x40\x02\x00\x40\x05\x04\x00\x00\x00\x64\xc0\x10\x10\x03\x0c\x00\x00\x00\x00\x00\x08\x00\x02\x03\xe8\x00\x00\x00\x02\x90\x0e\x00\x24\x00\x19\x46\x04\x01\x01\x01\x02\x00\x01\x19\x00\x01\x01\x01\x01\x02\x00\x02\x05\x00\x00\x03\xe8\x00\x00\x04\x00\x00\x00\x00\x00\x02\x00\x00\x02'
-__bgp6 = b'\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x00\x6f\x02\x00\x00\x00\x58\x40\x01\x01\x00\x40\x02\x00\x40\x05\x04\x00\x00\x00\x64\xc0\x10\x10\x03\x0c\x00\x00\x00\x00\x00\x08\x00\x02\x03\xe8\x00\x00\x00\x02\x90\x0e\x00\x33\x00\x19\x46\x04\x01\x01\x01\x02\x00\x02\x28\x00\x01\x01\x01\x01\x02\x00\x02\x05\x00\x00\x03\xe8\x00\x00\x04\x00\x00\x00\x00\x00\x02\x30\xcc\xaa\x02\x9c\xd8\x29\x20\xc0\xb4\x01\x02\x00\x00\x02\x00\x00\x00'
-__bgp7 = b'\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x00\x58\x02\x00\x00\x00\x41\x40\x01\x01\x00\x40\x02\x00\x40\x05\x04\x00\x00\x00\x64\xc0\x10\x10\x03\x0c\x00\x00\x00\x00\x00\x08\x00\x02\x03\xe8\x00\x00\x00\x02\x90\x0e\x00\x1c\x00\x19\x46\x04\x01\x01\x01\x02\x00\x03\x11\x00\x01\x01\x01\x01\x02\x00\x02\x00\x00\x00\x02\x20\xc0\xb4\x01\x02'
-__bgp8 = b'\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x00\x5f\x02\x00\x00\x00\x48\x40\x01\x01\x00\x40\x02\x00\x40\x05\x04\x00\x00\x00\x64\xc0\x10\x10\x03\x0c\x00\x00\x00\x00\x00\x08\x00\x02\x03\xe8\x00\x00\x00\x02\x90\x0e\x00\x23\x00\x19\x46\x04\x01\x01\x01\x02\x00\x04\x18\x00\x01\x01\x01\x01\x02\x00\x02\x05\x00\x00\x03\xe8\x00\x00\x04\x00\x00\x20\xc0\xb4\x01\x02'
-__bgp9 = b'\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x00\x7b\x02\x00\x00\x00\x64\x40\x01\x01\x00\x40\x02\x00\x40\x05\x04\x00\x00\x00\x64\xc0\x10\x10\x03\x0c\x00\x00\x00\x00\x00\x08\x00\x02\x03\xe8\x00\x00\x00\x02\x90\x0e\x00\x3f\x00\x19\x46\x04\x01\x01\x01\x02\x00\x02\x34\x00\x01\x01\x01\x01\x02\x00\x02\x05\x00\x00\x03\xe8\x00\x00\x04\x00\x00\x00\x00\x00\x02\x30\xcc\xaa\x02\x9c\xd8\x29\x80\xc0\xb4\x01\x02\xc0\xb4\x01\x02\xc0\xb4\x01\x02\xc0\xb4\x01\x02\x00\x00\x02\x00\x00\x00'
+__bgp5 = (
+    b'\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x00\x60\x02\x00\x00\x00\x49\x40\x01'
+    b'\x01\x00\x40\x02\x00\x40\x05\x04\x00\x00\x00\x64\xc0\x10\x10\x03\x0c\x00\x00\x00\x00\x00\x08\x00\x02'
+    b'\x03\xe8\x00\x00\x00\x02\x90\x0e\x00\x24\x00\x19\x46\x04\x01\x01\x01\x02\x00\x01\x19\x00\x01\x01\x01'
+    b'\x01\x02\x00\x02\x05\x00\x00\x03\xe8\x00\x00\x04\x00\x00\x00\x00\x00\x02\x00\x00\x02'
+)
+__bgp6 = (
+    b'\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x00\x6f\x02\x00\x00\x00\x58\x40\x01'
+    b'\x01\x00\x40\x02\x00\x40\x05\x04\x00\x00\x00\x64\xc0\x10\x10\x03\x0c\x00\x00\x00\x00\x00\x08\x00\x02'
+    b'\x03\xe8\x00\x00\x00\x02\x90\x0e\x00\x33\x00\x19\x46\x04\x01\x01\x01\x02\x00\x02\x28\x00\x01\x01\x01'
+    b'\x01\x02\x00\x02\x05\x00\x00\x03\xe8\x00\x00\x04\x00\x00\x00\x00\x00\x02\x30\xcc\xaa\x02\x9c\xd8\x29'
+    b'\x20\xc0\xb4\x01\x02\x00\x00\x02\x00\x00\x00'
+)
+__bgp7 = (
+    b'\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x00\x58\x02\x00\x00\x00\x41\x40\x01'
+    b'\x01\x00\x40\x02\x00\x40\x05\x04\x00\x00\x00\x64\xc0\x10\x10\x03\x0c\x00\x00\x00\x00\x00\x08\x00\x02'
+    b'\x03\xe8\x00\x00\x00\x02\x90\x0e\x00\x1c\x00\x19\x46\x04\x01\x01\x01\x02\x00\x03\x11\x00\x01\x01\x01'
+    b'\x01\x02\x00\x02\x00\x00\x00\x02\x20\xc0\xb4\x01\x02'
+)
+__bgp8 = (
+    b'\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x00\x5f\x02\x00\x00\x00\x48\x40\x01'
+    b'\x01\x00\x40\x02\x00\x40\x05\x04\x00\x00\x00\x64\xc0\x10\x10\x03\x0c\x00\x00\x00\x00\x00\x08\x00\x02'
+    b'\x03\xe8\x00\x00\x00\x02\x90\x0e\x00\x23\x00\x19\x46\x04\x01\x01\x01\x02\x00\x04\x18\x00\x01\x01\x01'
+    b'\x01\x02\x00\x02\x05\x00\x00\x03\xe8\x00\x00\x04\x00\x00\x20\xc0\xb4\x01\x02'
+)
+__bgp9 = (
+    b'\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x00\x7b\x02\x00\x00\x00\x64\x40\x01'
+    b'\x01\x00\x40\x02\x00\x40\x05\x04\x00\x00\x00\x64\xc0\x10\x10\x03\x0c\x00\x00\x00\x00\x00\x08\x00\x02'
+    b'\x03\xe8\x00\x00\x00\x02\x90\x0e\x00\x3f\x00\x19\x46\x04\x01\x01\x01\x02\x00\x02\x34\x00\x01\x01\x01'
+    b'\x01\x02\x00\x02\x05\x00\x00\x03\xe8\x00\x00\x04\x00\x00\x00\x00\x00\x02\x30\xcc\xaa\x02\x9c\xd8\x29'
+    b'\x80\xc0\xb4\x01\x02\xc0\xb4\x01\x02\xc0\xb4\x01\x02\xc0\xb4\x01\x02\x00\x00\x02\x00\x00\x00'
+)
 
 
 def test_pack():
@@ -798,19 +893,19 @@ def test_unpack():
     assert (len(b3.update.announced) == 0)
     assert (len(b3.update.attributes) == 6)
     a = b3.update.attributes[0]
-    assert (a.optional == False)
-    assert (a.transitive == True)
-    assert (a.partial == False)
-    assert (a.extended_length == False)
+    assert (not a.optional)
+    assert (a.transitive)
+    assert (not a.partial)
+    assert (not a.extended_length)
     assert (a.type == ORIGIN)
     assert (a.len == 1)
     o = a.origin
     assert (o.type == ORIGIN_IGP)
     a = b3.update.attributes[5]
-    assert (a.optional == True)
-    assert (a.transitive == False)
-    assert (a.partial == False)
-    assert (a.extended_length == True)
+    assert (a.optional)
+    assert (not a.transitive)
+    assert (not a.partial)
+    assert (a.extended_length)
     assert (a.type == MP_REACH_NLRI)
     assert (a.len == 30)
     m = a.mp_reach_nlri
@@ -936,7 +1031,553 @@ def test_unpack():
     assert (r.mpls_label_stack == b'\x00\x00\x02\x00\x00\x00')
 
 
-if __name__ == '__main__':
-    test_pack()
-    test_unpack()
-    print('Tests Successful...')
+def test_bgp_mp_nlri_20_1_mp_reach_nlri_next_hop():
+    # test for https://github.com/kbandla/dpkt/issues/485
+    __bgp = (
+        b'\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x00\x6c\x02\x00\x00\x00\x55\x40\x01'
+        b'\x01\x00\x40\x02\x04\x02\x01\xfd\xe9\x80\x04\x04\x00\x00\x00\x00\x80\x0e\x40\x00\x02\x01\x20\x20\x01'
+        b'\x0d\xb8\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\xfe\x80\x00\x00\x00\x00\x00\x00\xc0\x01\x0b'
+        b'\xff\xfe\x7e\x00\x00\x00\x40\x20\x01\x0d\xb8\x00\x01\x00\x02\x40\x20\x01\x0d\xb8\x00\x01\x00\x01\x40'
+        b'\x20\x01\x0d\xb8\x00\x01\x00\x00'
+    )
+    assert (__bgp == bytes(BGP(__bgp)))
+    bgp = BGP(__bgp)
+    assert (len(bgp.data) == 89)
+    assert (bgp.type == UPDATE)
+    assert (len(bgp.update.withdrawn) == 0)
+    assert (len(bgp.update.announced) == 0)
+    assert (len(bgp.update.attributes) == 4)
+
+    attribute = bgp.update.attributes[0]
+    assert (attribute.type == ORIGIN)
+    assert (not attribute.optional)
+    assert (attribute.transitive)
+    assert (not attribute.partial)
+    assert (not attribute.extended_length)
+    assert (attribute.len == 1)
+    o = attribute.origin
+    assert (o.type == ORIGIN_IGP)
+
+    attribute = bgp.update.attributes[1]
+    assert (attribute.type == AS_PATH)
+    assert (not attribute.optional)
+    assert (attribute.transitive)
+    assert (not attribute.partial)
+    assert (not attribute.extended_length)
+    assert (attribute.flags == 64)
+    assert (attribute.len == 4)
+    assert (len(attribute.as_path.segments) == 1)
+    segment = attribute.as_path.segments[0]
+    assert (segment.type == AS_SEQUENCE)
+    assert (segment.len == 1)
+    assert (len(segment.path) == 1)
+    assert (segment.path[0] == 65001)
+
+    attribute = bgp.update.attributes[2]
+    assert (attribute.type == MULTI_EXIT_DISC)
+    assert (attribute.optional)
+    assert (not attribute.transitive)
+    assert (not attribute.partial)
+    assert (not attribute.extended_length)
+    assert (attribute.flags == 0x80)
+    assert (attribute.len == 4)
+    assert (attribute.multi_exit_disc.value == 0)
+
+    attribute = bgp.update.attributes[3]
+    assert (attribute.type == MP_REACH_NLRI)
+    assert (attribute.optional)
+    assert (not attribute.transitive)
+    assert (not attribute.partial)
+    assert (not attribute.extended_length)
+    assert (attribute.flags == 0x80)
+    assert (attribute.len == 64)
+    mp_reach_nlri = attribute.mp_reach_nlri
+    assert (mp_reach_nlri.afi == AFI_IPV6)
+    assert (mp_reach_nlri.safi == SAFI_UNICAST)
+    assert (len(mp_reach_nlri.snpas) == 0)
+    assert (len(mp_reach_nlri.announced) == 3)
+    prefix = mp_reach_nlri.announced[0]
+    assert (socket.inet_ntop(socket.AF_INET6, prefix.prefix) == '2001:db8:1:2::')
+    assert (prefix.len == 64)
+    prefix = mp_reach_nlri.announced[1]
+    assert (socket.inet_ntop(socket.AF_INET6, prefix.prefix) == '2001:db8:1:1::')
+    assert (prefix.len == 64)
+    prefix = mp_reach_nlri.announced[2]
+    assert (socket.inet_ntop(socket.AF_INET6, prefix.prefix) == '2001:db8:1::')
+    assert (prefix.len == 64)
+    assert (len(mp_reach_nlri.next_hops) == 2)
+    assert (socket.inet_ntop(socket.AF_INET6, mp_reach_nlri.next_hops[0]) == '2001:db8::1')
+    assert (socket.inet_ntop(socket.AF_INET6, mp_reach_nlri.next_hops[1]) == 'fe80::c001:bff:fe7e:0')
+    assert (mp_reach_nlri.next_hop == b''.join(mp_reach_nlri.next_hops))
+
+
+def test_bgp_add_path_6_1_as_path():
+    # test for https://github.com/kbandla/dpkt/issues/481
+    # Error processing BGP data: packet 6 : message 1 of bgp-add-path.cap
+    # https://packetlife.net/media/captures/bgp-add-path.cap
+    __bgp = (
+        b'\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x00\x59\x02\x00\x00\x00\x30\x40\x01'
+        b'\x01\x00\x40\x02\x06\x02\x01\x00\x00\xfb\xff\x40\x03\x04\x0a\x00\x0e\x01\x80\x04\x04\x00\x00\x00\x00'
+        b'\x40\x05\x04\x00\x00\x00\x64\x80\x0a\x04\x0a\x00\x22\x04\x80\x09\x04\x0a\x00\x0f\x01\x00\x00\x00\x01'
+        b'\x20\x05\x05\x05\x05\x00\x00\x00\x01\x20\xc0\xa8\x01\x05'
+    )
+    bgp = BGP(__bgp)
+    assert (__bgp == bytes(bgp))
+    assert (len(bgp) == 89)
+    assert (bgp.type == UPDATE)
+    assert (len(bgp.update.withdrawn) == 0)
+    announced = bgp.update.announced
+    assert (len(announced) == 2)
+    assert (announced[0].len == 32)
+    assert (announced[0].path_id == 1)
+    assert (socket.inet_ntop(socket.AF_INET, bytes(announced[0].prefix)) == '5.5.5.5')
+    assert (announced[1].len == 32)
+    assert (announced[1].path_id == 1)
+    assert (socket.inet_ntop(socket.AF_INET, bytes(announced[1].prefix)) == '192.168.1.5')
+
+    assert (len(bgp.update.attributes) == 7)
+
+    attribute = bgp.update.attributes[0]
+    assert (attribute.type == ORIGIN)
+    assert (not attribute.optional)
+    assert (attribute.transitive)
+    assert (not attribute.partial)
+    assert (not attribute.extended_length)
+    assert (attribute.flags == 0x40)
+    assert (attribute.len == 1)
+    assert (attribute.origin.type == ORIGIN_IGP)
+
+    attribute = bgp.update.attributes[1]
+    assert (attribute.type == AS_PATH)
+    assert (not attribute.optional)
+    assert (attribute.transitive)
+    assert (not attribute.partial)
+    assert (not attribute.extended_length)
+    assert (attribute.flags == 0x40)
+    assert (attribute.len == 6)
+    assert (len(attribute.as_path.segments) == 1)
+    segment = attribute.as_path.segments[0]
+    assert (segment.type == AS_SEQUENCE)
+    assert (segment.len == 1)
+    assert (len(segment.path) == 1)
+    assert (segment.path[0] == 64511)
+
+    attribute = bgp.update.attributes[2]
+    assert (attribute.type == NEXT_HOP)
+    assert (not attribute.optional)
+    assert (attribute.transitive)
+    assert (not attribute.partial)
+    assert (not attribute.extended_length)
+    assert (attribute.flags == 0x40)
+    assert (attribute.len == 4)
+    assert (socket.inet_ntop(socket.AF_INET, bytes(attribute.next_hop)) == '10.0.14.1')
+
+    attribute = bgp.update.attributes[3]
+    assert (attribute.type == MULTI_EXIT_DISC)
+    assert (attribute.optional)
+    assert (not attribute.transitive)
+    assert (not attribute.partial)
+    assert (not attribute.extended_length)
+    assert (attribute.flags == 0x80)
+    assert (attribute.len == 4)
+    assert (attribute.multi_exit_disc.value == 0)
+
+    attribute = bgp.update.attributes[4]
+    assert (attribute.type == LOCAL_PREF)
+    assert (not attribute.optional)
+    assert (attribute.transitive)
+    assert (not attribute.partial)
+    assert (not attribute.extended_length)
+    assert (attribute.flags == 0x40)
+    assert (attribute.len == 4)
+    assert (attribute.local_pref.value == 100)
+
+    attribute = bgp.update.attributes[5]
+    assert (attribute.type == CLUSTER_LIST)
+    assert (attribute.optional)
+    assert (not attribute.transitive)
+    assert (not attribute.partial)
+    assert (not attribute.extended_length)
+    assert (attribute.flags == 0x80)
+    assert (attribute.len == 4)
+    assert (socket.inet_ntop(socket.AF_INET, bytes(attribute.cluster_list)) == '10.0.34.4')
+
+    attribute = bgp.update.attributes[6]
+    assert (attribute.type == ORIGINATOR_ID)
+    assert (attribute.optional)
+    assert (not attribute.transitive)
+    assert (not attribute.partial)
+    assert (not attribute.extended_length)
+    assert (attribute.flags == 0x80)
+    assert (attribute.len == 4)
+    assert (socket.inet_ntop(socket.AF_INET, bytes(attribute.originator_id)) == '10.0.15.1')
+
+
+def test_attribute_accessors():
+    from binascii import unhexlify
+
+    buf = unhexlify(
+        '00'  # flags
+        '01'  # type (ORIGIN)
+
+        '01'  # length
+        '00'  # Origin type
+    )
+    attribute = BGP.Update.Attribute(buf)
+    assert isinstance(attribute.data, BGP.Update.Attribute.Origin)
+    for attr in ['optional', 'transitive', 'partial', 'extended_length']:
+        assert getattr(attribute, attr) == 0
+
+        # check we can set..
+        setattr(attribute, attr, 1)
+        assert getattr(attribute, attr) == 1
+
+        # and also unset
+        setattr(attribute, attr, 0)
+        assert getattr(attribute, attr) == 0
+
+
+def test_snpa():
+    from binascii import unhexlify
+
+    buf = unhexlify(
+        '04'    # len (in semi-octets)
+        '1234'  # data
+    )
+    snpa = BGP.Update.Attribute.MPReachNLRI.SNPA(buf)
+    assert snpa.len == 4  # length of the data in semi-octets
+    assert len(snpa) == 3  # length of the snpa in bytes (including header)
+    assert bytes(snpa) == buf
+
+
+def test_mpreachnlri():
+    from binascii import unhexlify
+
+    buf = unhexlify(
+        '0000'  # afi
+        '00'    # safi
+
+        '00'    # nlen
+        '01'    # num SNPAs
+
+        # SNPA
+        '04'    # len
+        '1234'  # data
+    )
+    mp = BGP.Update.Attribute.MPReachNLRI(buf)
+    assert len(mp.snpas) == 1
+    assert bytes(mp) == buf
+
+
+def test_notification():
+    from binascii import unhexlify
+
+    buf_notification = unhexlify(
+        '11'   # code
+        '22'   # subcode
+
+        '33'   # error
+    )
+    notification = BGP.Notification(buf_notification)
+    assert notification.code == 0x11
+    assert notification.subcode == 0x22
+    assert notification.error == b'\x33'
+    assert bytes(notification) == buf_notification
+
+    buf_bgp_hdr = unhexlify(
+        '11111111111111111111111111111111'  # marker
+        '0016'  # len
+        '03'    # type (NOTIFICATION)
+    )
+    bgp = BGP(buf_bgp_hdr + buf_notification)
+
+    assert hasattr(bgp, 'notification')
+    assert isinstance(bgp.data, BGP.Notification)
+    assert bgp.data.code == 0x11
+    assert bgp.data.subcode == 0x22
+    assert bgp.data.error == b'\x33'
+    assert bytes(bgp) == buf_bgp_hdr + buf_notification
+
+
+def test_keepalive():
+    keepalive = BGP.Keepalive(b'\x11')
+    assert len(keepalive) == 0
+    assert bytes(keepalive) == b''
+
+
+def test_routegeneric():
+    from binascii import unhexlify
+
+    buf = unhexlify(
+        '08'  # len (bits)
+        '11'  # prefix
+    )
+    routegeneric = RouteGeneric(buf)
+    assert routegeneric.len == 8
+    assert routegeneric.prefix == b'\x11'
+
+    assert bytes(routegeneric) == buf
+    assert len(routegeneric) == 2
+
+
+def test_routeipv4():
+    from binascii import unhexlify
+
+    buf = unhexlify(
+        '08'  # len (bits)
+
+        '11'  # prefix
+    )
+    routeipv4 = RouteIPV4(buf)
+    assert routeipv4.len == 8  # prefix len in bits
+    assert routeipv4.prefix == b'\x11\x00\x00\x00'
+
+    assert repr(routeipv4) == "RouteIPV4(17.0.0.0/8)"
+    assert bytes(routeipv4) == buf
+    assert len(routeipv4) == 2  # header + prefix(bytes)
+
+
+def test_routeipv6():
+    from binascii import unhexlify
+
+    buf = unhexlify(
+        '08'  # len (bits)
+        '22'  # prefix
+    )
+
+    routeipv6 = RouteIPV4(buf)
+    assert routeipv6.len == 8  # prefix len in bits
+    assert routeipv6.prefix == b'\x22\x00\x00\x00'
+
+    assert bytes(routeipv6) == buf
+    assert len(routeipv6) == 2  # header + prefix(bytes)
+
+
+def test_extendedrouteipv4():
+    from binascii import unhexlify
+
+    buf = unhexlify(
+        '00000001'  # path_id
+        '20'        # len (bits)
+        '05050505'  # prefix
+    )
+    extendedrouteipv4 = ExtendedRouteIPV4(buf)
+    assert extendedrouteipv4.path_id == 1
+    assert extendedrouteipv4.len == 32
+    assert extendedrouteipv4.prefix == unhexlify('05050505')
+    assert repr(extendedrouteipv4) == "ExtendedRouteIPV4(5.5.5.5/32 PathId 1)"
+
+    assert bytes(extendedrouteipv4) == buf
+    assert len(extendedrouteipv4) == len(buf)
+
+
+def test_routeevpn():
+    from binascii import unhexlify
+
+    buf = unhexlify(
+        '02'  # type
+        '1a'  # len
+
+        # route distinguisher
+        '1111111111111111'
+
+        # esi
+        '22222222222222222222'
+
+        # eth_id
+        '33333333'
+
+        # mac address
+        '00'  # len (bits)
+
+        # ip address
+        '00'  # len (bits)
+
+        # mpls
+        '6666'  # label stack
+    )
+
+    routeevpn = RouteEVPN(buf)
+    assert routeevpn.type == 2
+    assert routeevpn.len == 26
+
+    assert routeevpn.esi == unhexlify('22222222222222222222')
+    assert routeevpn.eth_id == unhexlify('33333333')
+
+    assert routeevpn.mac_address_length == 0
+    assert routeevpn.mac_address is None
+
+    assert routeevpn.ip_address_length == 0
+    assert routeevpn.ip_address is None
+
+    assert routeevpn.mpls_label_stack == unhexlify('6666')
+
+    assert bytes(routeevpn) == buf
+    assert len(routeevpn) == len(buf)
+
+
+def test_route_refresh():
+    from binascii import unhexlify
+    buf_route_refresh = unhexlify(
+        '1111'  # afi
+        '22'    # rsvd
+        '33'    # safi
+    )
+    route_refresh = BGP.RouteRefresh(buf_route_refresh)
+    assert route_refresh.afi == 0x1111
+    assert route_refresh.rsvd == 0x22
+    assert route_refresh.safi == 0x33
+    assert bytes(route_refresh) == buf_route_refresh
+
+    buf_bgp_hdr = unhexlify(
+        '11111111111111111111111111111111'  # marker
+        '0017'  # len
+        '05'    # type (ROUTE_REFRESH)
+    )
+    bgp = BGP(buf_bgp_hdr + buf_route_refresh)
+
+    assert hasattr(bgp, 'route_refresh')
+    assert isinstance(bgp.data, BGP.RouteRefresh)
+    assert bgp.data.afi == 0x1111
+    assert bgp.data.rsvd == 0x22
+    assert bgp.data.safi == 0x33
+    assert bytes(bgp) == buf_bgp_hdr + buf_route_refresh
+
+
+def test_mpunreachnlri():
+    from binascii import unhexlify
+    buf_routeipv4 = unhexlify(
+        '08'  # len (bits)
+        '11'  # prefix
+    )
+
+    buf_routeipv6 = unhexlify(
+        '08'  # len (bits)
+        '22'  # prefix
+    )
+
+    buf_routeevpn = unhexlify(
+        '02'  # type
+        '1a'  # len
+
+        # route distinguisher
+        '1111111111111111'
+
+        # esi
+        '22222222222222222222'
+
+        # eth_id
+        '33333333'
+
+        # mac address
+        '00'  # len (bits)
+
+        # ip address
+        '00'  # len (bits)
+
+        # mpls
+        '6666'  # label stack
+    )
+
+    buf_routegeneric = unhexlify(
+        '08'  # len (bits)
+        '33'  # prefix
+    )
+
+    afi = struct.Struct('>H')
+    routes = (
+        (AFI_IPV4, buf_routeipv4, RouteIPV4),
+        (AFI_IPV6, buf_routeipv6, RouteIPV6),
+        (AFI_L2VPN, buf_routeevpn, RouteEVPN),
+        # this afi does not exist, so we will parse as RouteGeneric
+        (1234, buf_routegeneric, RouteGeneric),
+    )
+
+    for afi_id, buf, cls in routes:
+        buf = afi.pack(afi_id) + b'\xcc' + buf
+        mpu = BGP.Update.Attribute.MPUnreachNLRI(buf)
+
+        assert mpu.afi == afi_id
+        assert mpu.safi == 0xcc
+        assert len(mpu.data) == 1
+        route = mpu.data[0]
+        assert isinstance(route, cls)
+
+        assert bytes(mpu) == buf
+        assert len(mpu) == len(buf)
+
+    # test the unpacking of the routes, as an Attribute
+    attribute_hdr = struct.Struct('BBB')
+    for afi_id, buf, cls in routes:
+        buf_mpunreachnlri = afi.pack(afi_id) + b'\xcc' + buf
+        buf_attribute_hdr = attribute_hdr.pack(0, MP_UNREACH_NLRI, len(buf_mpunreachnlri))
+        buf = buf_attribute_hdr + buf_mpunreachnlri
+
+        attribute = BGP.Update.Attribute(buf)
+        assert isinstance(attribute.data, BGP.Update.Attribute.MPUnreachNLRI)
+        routes = attribute.data.data
+        assert len(routes) == 1
+        assert isinstance(routes[0], cls)
+
+
+def test_update_withdrawn():
+    from binascii import unhexlify
+    buf_ipv4 = unhexlify(
+        '08'  # len (bits)
+        '11'  # prefix
+    )
+    packed_length = struct.Struct('>H').pack
+    wlen, plen = packed_length(len(buf_ipv4)), packed_length(0)
+
+    buf = wlen + buf_ipv4 + plen
+    update = BGP.Update(buf)
+
+    assert len(update.withdrawn) == 1
+    route = update.withdrawn[0]
+    assert isinstance(route, RouteIPV4)
+    assert bytes(update) == buf
+
+
+def test_parameters():
+    from binascii import unhexlify
+    buf = unhexlify(
+        '44'        # v
+        '1111'      # asn
+        '2222'      # holdtime
+        '33333333'  # identifier
+        '03'        # param_len
+
+        # Parameter
+        '01'  # type (AUTHENTICATION)
+        '01'  # len
+
+        # Authentication
+        '11'  # code
+    )
+    bgp_open = BGP.Open(buf)
+    assert len(bgp_open.parameters) == 1
+    parameter = bgp_open.parameters[0]
+
+    assert isinstance(parameter, BGP.Open.Parameter)
+    assert isinstance(parameter.data, BGP.Open.Parameter.Authentication)
+
+    assert bytes(bgp_open) == buf
+    assert len(bgp_open) == len(buf)
+
+
+def test_reservedcommunities():
+    from binascii import unhexlify
+    buf = unhexlify(
+        # ReservedCommunity
+        '00002222'  # value
+    )
+    communities = BGP.Update.Attribute.Communities(buf)
+    assert len(communities.data) == 1
+
+    community = communities.data[0]
+    assert isinstance(community, BGP.Update.Attribute.Communities.ReservedCommunity)
+    assert len(community) == 4
+    assert bytes(community) == buf
+
+    assert len(communities) == 4
+    assert bytes(communities) == buf
